""" Global registry for available event interfaces

Intended usage:

    Various generic Event processors may want to be able to list possible
    events, such as for listing them for selection in a UI.

    - Event interfaces should be registered with this registry through the
      registerEvent call, passing in both the interface and a title, and
      optionally a definition. If no definition is given, the Interface
      docstring will be used instead.

    - Consumers interested in event interfaces can use the listEvents and
      getEvent* methods to discover available Event information.

$Id: EventRegistry.py,v 1.1 2003/03/12 10:57:56 smigniot Exp $
"""
from zLOG import LOG, WARNING

_registry = {}


class EventRegistry:
    """Registry for Event Interfaces.
    
    Modules make Event interfaces available for listing by registering them
    here.

    """

    def listEvents(self, with_metadata=0):
        """List the registered event interfaces.
        
        - If 'with_metadata', then return a sequence of (key, title,
          description) tuples.

        - Otherwise, return a sequence of keys.

        """
        if not with_metadata:
            return _registry.keys()

        return [(k, v['title'], v['description'])
                for k, v in _registry.items()]

    def getEvent(self, key):
        """Look up the event information.

        Returns a dict with title, description and interface keys.

        """
        return _registry.get(key)

    def getEventTitle(self, key):
        """Look up the title for a given event key."""
        data = _registry.get(key)
        return data and data['title']

    def getEventDescription(self, key):
        """Look up the description for a given event key."""
        data = _registry.get(key)
        return data and data['description']

    def getEventInterface(self, key):
        """Look up the event interface for a given event key."""
        data = _registry.get(key)
        return data and data['interface']

    def registerEvent(self, interface, title, description=None):
        """Register an event interface.

        'interface'     the event interface to register
        'title'         the title of the interface
        'description'   documentation for the event. If not given, the
                        docstring of the event interface is used instead.

        Returns the key that will be used to access the event.
        """
        key = '%s.%s' % (interface.__module__, interface.__name__)
        if _registry.has_key( key ):
            LOG('EventRegistry', WARNING,
                'Duplicate registration for event: %s' % key)
            return key

        if description is None:
            description = reindentDocstring(interface.__doc__)

        _registry[key] = {'interface': interface, 'title': title,
                          'description': description}


        return key

eventRegistry = EventRegistry()


def reindentDocstring(doc):
    """Remove indentation from docstring.

    It is assumed that the first line is not indented, and that the first
    non-blank line after that can be used to determine indentation for the
    whole docstring, as is the case with this docstring. This will help
    preserve indentation within the docstring itself.

    """

    lines = doc.splitlines()

    if len(lines) == 1:
        return doc

    indent = None
    for i in range(1, len(lines)):
        if indent is None and lines[i].strip():
            indent = len(lines[i]) - len(lines[i].lstrip())
            lines[i] = lines[i][indent:]
        elif indent:
            lines[i] = lines[i][indent:]

    return "\n".join(lines)
