import unittest

class DummyEvent:
    """Dummy Event semi-interface."""

class EventRegistryTests(unittest.TestCase):
    oldRegistry = None

    def setUp(self):
        self._clearRegistry()

    def tearDown(self):
        self._clearRegistry()

    def _clearRegistry(self):
        from Products.Event.EventRegistry import _registry
        if self.oldRegistry is None:
            self.oldRegistry = _registry.copy()
            _registry.clear()
        else:
            _registry.clear()
            _registry.update(self.oldRegistry)
        
    def _makeOne(self):
        from Products.Event import eventRegistry
        return eventRegistry

    def test_empty(self):
        registry = self._makeOne()

        self.assertEqual(len(registry.listEvents()), 0)
        self.assertEqual(registry.getEvent('foo'), None)


    def test_registerEvent(self):
        registry = self._makeOne()
        registry.registerEvent(DummyEvent, 'Dummy Event')
        key = '%s.%s' % (DummyEvent.__module__, DummyEvent.__name__)

        self.assertEqual(len(registry.listEvents()), 1)
        self.failUnless(key in registry.listEvents())
        self.failUnless((key, 'Dummy Event', DummyEvent.__doc__)
                        in registry.listEvents(with_metadata=1))

        metadata = registry.getEvent(key)
        self.assertEqual(metadata['title'], 'Dummy Event')
        self.assertEqual(metadata['description'], DummyEvent.__doc__)
        self.failUnless(metadata['interface'] is DummyEvent)

        class Bar(DummyEvent): pass

        registry.registerEvent(Bar, 'Dummy Event (bar)', 'Bar Description')
        key = '%s.%s' % (Bar.__module__, Bar.__name__)

        self.assertEqual(len(registry.listEvents()), 2)
        self.failUnless(key in registry.listEvents())
        self.assertEqual(registry.getEventTitle(key), 'Dummy Event (bar)')
        self.assertEqual(registry.getEventDescription(key), 'Bar Description')
        self.failUnless(registry.getEventInterface(key) is Bar)

class ReindentDocstringTest(unittest.TestCase):
    def _reindent(self, doc):
        from Products.Event.EventRegistry import reindentDocstring
        return reindentDocstring(doc)

    def testOneLine(self):
        doc = """This is a online docstring."""
        self.assertEqual(self._reindent(doc), doc)

    def testNoIndent(self):
        doc = "No indentation.\n\nBut multiple lines nonetheless."
        self.assertEqual(self._reindent(doc), doc)

    def testOneLevelIndent(self):
        doc = """This is a multiline docstring.

        All lines are at the same indentation
        except for the first line.

        And it contains several blank lines.

        """

        self.assertEqual(self._reindent(doc), "\n".join((
            "This is a multiline docstring.",
            "",
            "All lines are at the same indentation",
            "except for the first line.",
            "",
            "And it contains several blank lines.",
            "", "")))

    def testMultiLevelIndent(self):
        doc = """This is a more complex multiline docstring.

        In this docstring, multiple levels of indent are used:

        - With bullets across a line break
          indentation keeps the bullet together.

          - Sub bullets use further indentation

          - Code examples are preserved:

            def foo(bar):
                return 'Spam and eggs'

        """

        self.assertEqual(self._reindent(doc), "\n".join((
            "This is a more complex multiline docstring.",
            "",
            "In this docstring, multiple levels of indent are used:",
            "",
            "- With bullets across a line break",
            "  indentation keeps the bullet together.",
            "",
            "  - Sub bullets use further indentation",
            "",
            "  - Code examples are preserved:",
            "",
            "    def foo(bar):",
            "        return 'Spam and eggs'",
            "", "")))


def test_suite():
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    for testcase in (EventRegistryTests, ReindentDocstringTest):
        suite.addTest(loader.loadTestsFromTestCase(testcase))
    return suite

if __name__ == '__main__':
    unittest.main()
