# Copyright (c) 2002 Nuxeo SARL <http://nuxeo.com>
# Copyright (c) 2002 Prfecture du Bas-Rhin, France
# Author: 2002 Florent Guillaume <mailto:fg@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: ReviewedHierarchyLevel.py,v 1.17 2003/07/14 21:30:14 fermigier Exp $

"""
  Reviewed Hierarchy Level
"""

__version__ = '$Revision: 1.17 $'[11:-2]

import string

from Globals import InitializeClass
from AccessControl import ClassSecurityInfo

from Products.ZCatalog.ZCatalog import ZCatalog

from Products.CMFCore.CMFCorePermissions import \
     View, AccessContentsInformation
from Products.CMFCore.utils import getToolByName

from publication_utils import modifyRolesForPermission

from HierarchiesTool import ChangeHierarchyLevel
from BaseHierarchyLevel import BaseHierarchyLevel


AnonymousRole = 'Anonymous'
ManagerRole = 'Manager'
ReaderRole = 'Reader'
ReviewerRole = 'Reviewer'


factory_type_information = (
    {'id': 'Reviewed Hierarchy Level',
     'description': ('A Reviewed Hierarchy Level belongs to a Hierarchy and '
                     'has a corresponding query into the catalog.'),
     'title': "Hierarchy",
     'content_icon': 'hierarchylevel_icon.gif',
     'product': 'NuxHierarchies',
     'meta_type': 'Reviewed Hierarchy Level',
     'factory': 'addReviewedHierarchyLevel',
     'immediate_view': 'reviewedhierarchylevel_edit_form',
     'filter_content_types': 1,
     'allowed_content_types': ('Reviewed Hierarchy Level',
                               ),
     'actions': ({'id': 'view',
                  'name': '_action_view_',
                  'action': 'reviewedhierarchylevel_view',
                  'permissions': (View,)},
                 {'id': 'edit',
                  'name': '_action_modify_',
                  'action': 'reviewedhierarchylevel_edit_form',
                  'permissions': (ChangeHierarchyLevel,)},
                 {'id': 'localroles',
                  'name': '_action_access_rights_',
                  'action': 'reviewedhierarchylevel_localrole_form',
                  'permissions': (ChangeHierarchyLevel,)},
                 {'id': 'folder_contents',
                  'name': '_action_subhierachies_',
                  'action': 'folder_contents',
                  'category': 'folder',
                  'permissions': (ChangeHierarchyLevel,)},
                 {'id': 'create',
                  'name': '_action_create_',
                  'action': 'reviewedhierarchylevel_create_form',
                  'visible': 0,
                  'permissions': ()},
                 ),
     },
    )


class ReviewedHierarchyLevel(BaseHierarchyLevel):
    """Reviewed Hierarchy Level

    This is a Skinned Folder which belongs to a Hierarchy and has a
    corresponding key.
    """
    meta_type = 'Reviewed Hierarchy Level'

    _properties = BaseHierarchyLevel._properties + (
        {'id': 'level_reviewers', 'type': 'lines', 'mode': 'w', 
            'label': 'Reviewers'},
        {'id': 'level_readers', 'type': 'lines', 'mode': 'w', 
            'label': 'Readers'},
        {'id': 'base_hierarchy', 'type': 'boolean', 'mode': 'w', 
            'label': 'Base hierarchy'},
        )

    level_reviewers = []
    level_readers = []

    base_hierarchy = 0

    security = ClassSecurityInfo()

    def __init__(self, id,
                 title='',
                 description='',
                 hierarchyid='main',
                 levelkey='',
                 publishable=1,
                 level_reviewers=[],
                 level_readers=[],
                 base_hierarchy=0):
        BaseHierarchyLevel.__init__(self, id,
                                    title=title,
                                    description=description,
                                    hierarchyid=hierarchyid,
                                    levelkey=levelkey,
                                    publishable=publishable)
        self.level_reviewers = list(level_reviewers)
        self.level_readers = list(level_readers)
        self.base_hierarchy = base_hierarchy


    #
    # Remove all documents in this level upon deletion
    #
    security.declarePrivate('manage_beforeDelete')
    def manage_beforeDelete(self, item, container):
        """Updates hierarchy upon deletion"""
        ReviewedHierarchyLevel.inheritedAttribute('manage_beforeDelete')(self, item, container)
        # Lookup all objects mentionning this level
        # and remove the level from them.
        key = self.getLevelKey()
        hier = self._get_hierarchy()
        hiername = hier.getId()
        brains = self.queryCatalog() + self.queryPendingCatalog()
        done = {}
        for brain in brains:
            u = brain.getURL()
            if done.has_key(u): 
                continue
            done[u] = None
            ob = brain.getObject()
            ob.removeCategory(key, hiername)

    #
    # Accessors
    #

    security.declarePublic('LevelReviewers')
    def LevelReviewers(self):
        """Returns the effective reviewers for this level, including acquired
        ones.
        """
        return tuple(self.level_reviewers)

    security.declarePublic('LevelReaders')
    def LevelReaders(self):
        """Returns the effective readers for this level, including acquired
        ones.
        """
        return tuple(self.level_readers)

    security.declarePublic('LevelLocalReviewers')
    def LevelLocalReviewers(self):
        """Returns the reviewers defined at this level."""
        return tuple(self.level_reviewers)

    security.declarePublic('LevelLocalReaders')
    def LevelLocalReaders(self):
        """Returns the readers defined at this level."""
        return tuple(self.level_readers)

    security.declarePublic('LevelIsPublic')
    def LevelIsPublic(self):
        """Returns true if the level is readable by all."""
        # XXX take acquisition into account
        return not self.level_readers

    #
    # Hierarchy internals
    #

    # called by MailingList
    security.declarePrivate('getContextUsers')
    def getContextUsers(self):
        return {
            'WorkgroupManager': self.LevelReviewers(),
            'WorkgroupMember': [],
            'WorkgroupVisitor': self.LevelReaders()
            }

    # called by BaseHierarchy
    security.declarePrivate('get_level_additional_info')
    def get_level_additional_info(self, dict):
        """Add to dict additional info."""
        dict['levelreviewers'] = self.LevelReviewers()
        dict['levelreaders'] = self.LevelReaders()

    # overload
    security.declareProtected(ChangeHierarchyLevel, 'edit')
    def edit(self,
             title=None,
             description=None,
             hierarchyid=None,
             levelkey=None,
             publishable=None,
             level_reviewers=None,
             level_readers=None,
             base_hierarchy=None,
             REQUEST=None, **kw):
        """Edit a Reviewed Hierarchy Level."""
        BaseHierarchyLevel.edit(self,
                                title=title,
                                description=description,
                                hierarchyid=hierarchyid,
                                levelkey=levelkey,
                                publishable=publishable)
        perm_changed = 0
        if level_reviewers is not None:
            level_reviewers = filter(None, map(string.strip, level_reviewers))
            self.level_reviewers = level_reviewers
            perm_changed = 1
        if level_readers is not None:
            level_readers = filter(None, map(string.strip, level_readers))
            self.level_readers = level_readers
            perm_changed = 1
        if base_hierarchy is not None:
            self.base_hierarchy = base_hierarchy
        if perm_changed:
            self.setPermissions()
            self.setDocumentsPermissions()
        # reindex for level reviewers
        self.reindexObject()
        # rebuild hierarchy XXX do it here ?
        self.rebuildHierarchy()
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect('%s/reviewedhierarchylevel_view' 
                % self.absolute_url())


    security.declarePrivate('setDocumentsPermissions')
    def setDocumentsPermissions(self):
        """Sets permissions on all documents mentionning this level
        so that the new readers or reviewers have adequate permissions.
        """
        # XXX because in the levels the readers should be are acquired,
        # we should do all sublevels too.
        hier = self._get_hierarchy()
        hiername = hier.getId()
        brains = self.queryCatalog() + self.queryPendingCatalog()
        done = {}
        for brain in brains:
            u = brain.getURL()
            if done.has_key(u): 
                continue
            done[u] = None
            ob = brain.getObject()
            ob.setWorkflowStatePermissions(hiername)


    security.declarePrivate('setPermissions')
    def setPermissions(self):
        """ Sets the permissions on the level for access by reviewers and
        readers only.

        NOTE !!! Because local roles are always acquired from the parent, this
        may end up giving more rights than you expect. Another filtering pass
        is needed before displaying a level.
        """
        #
        # Remove local roles Reader and Reviewer
        #
        # User local roles
        lroles = self.get_local_roles()
        for id, roles in lroles:
            roles = list(roles)
            changed = 0
            if ReaderRole in roles:
                roles.remove(ReaderRole)
                changed = 1
            if ReviewerRole in roles:
                roles.remove(ReviewerRole)
                changed = 1
            if changed:
                if roles:
                    self.manage_setLocalRoles(id, roles)
                else:
                    self.manage_delLocalRoles((id,))
        # Group local roles
        try:
            lgroups = self.get_local_group_roles()
        except AttributeError:
            lgroups = ()
        for id, roles in lgroups:
            roles = list(roles)
            changed = 0
            if ReaderRole in roles:
                roles.remove(ReaderRole)
                changed = 1
            if ReviewerRole in roles:
                roles.remove(ReviewerRole)
                changed = 1
            if changed:
                if roles:
                    self.manage_setLocalGroupRoles(id, roles)
                else:
                    self.manage_delLocalGroupRoles((id,))

        #
        # Set adequate local roles
        #
        reviewers = self.LevelLocalReviewers()
        readers = self.LevelLocalReaders()
        public = self.LevelIsPublic()
        for (ids, role) in ((reviewers, ReviewerRole), (readers, ReaderRole)):
            for id in ids:
                if not id.startswith('group:'):
                    self.manage_addLocalRoles(id, (role,))
                else:
                    id = id[len('group:'):]
                    try:
                        self.manage_addLocalGroupRoles(id, (role,))
                    except AttributeError:
                        pass

        # Set view permissions
        view = (ManagerRole, ReaderRole, ReviewerRole)
        if public:
            view = view + (AnonymousRole,)

        modifyRolesForPermission(self, View, view)
        # Don't change AccessContentsInformation, which is needed for
        # restrictedTraverse to function correctly.
        #modifyRolesForPermission(self, AccessContentsInformation, view)

        # reindex security
        self.reindexObjectSecurity()


    # overload
    security.declarePublic('publicationAllowedForUser')
    def publicationAllowedForUser(self, id=None):
        """Return true if the user (None=current) can publish to this level."""
        if not self.publishable:
            return 0
        if id is None:
            pm = getToolByName(self, 'portal_membership')
            id = pm.getAuthenticatedMember().getUserName()
        return id in self.LevelReviewers()

    #
    # Queries
    #

    security.declareProtected(AccessContentsInformation, 'buildPendingQuery')
    def buildPendingQuery(self):
        """Builds the query for the pending documents at this level."""
        levelkey = self.levelkey
        if not levelkey: 
            return {}
        criterion = self._get_hierarchy().criterion_pending
        return {criterion: levelkey}

    security.declareProtected(AccessContentsInformation, 'queryPendingCatalog')
    def queryPendingCatalog(self, REQUEST=None, **kw):
        """Queries the Catalog to get this level's pending documents."""
        query = self.buildPendingQuery()
        if not query: 
            return ()
        kw.update(query)
        portal_catalog = getToolByName(self, 'portal_catalog')
        return portal_catalog.searchResults(REQUEST, **kw)

    security.declarePrivate('queryPendingCatalogNoSecurity')
    def queryPendingCatalogNoSecurity(self, REQUEST=None, **kw):
        """Queries the Catalog to get this level's pending documents.

        Accesses the ZCatalog directly to bypass allowedRolesAndUsers
        filtering. This is needed to do a query during traversal where
        the user is not yet authenticated.
        """
        query = self.buildPendingQuery()
        if not query: 
            return ()
        kw.update(query)
        portal_catalog = getToolByName(self, 'portal_catalog')
        return ZCatalog.searchResults(portal_catalog, REQUEST, **kw)


InitializeClass(ReviewedHierarchyLevel)


def addReviewedHierarchyLevel(dispatcher, id, REQUEST=None):
    """Add a Reviewed Hierarchy Level."""
    ob = ReviewedHierarchyLevel(id)
    container = dispatcher.Destination()
    container._setObject(id, ob)
    ob = container._getOb(id)
    # set default hierarchyid to its parent's
    if hasattr(container, 'getProperty'):
        hierarchyid = container.getProperty('hierarchyid', None)
        if hierarchyid is not None and type(hierarchyid) is type(''):
            ob.hierarchyid = hierarchyid
    if REQUEST is not None:
        url = dispatcher.DestinationURL()
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)

