# Copyright (C) 2002 Nuxeo SARL <http://nuxeo.com>
# Copyright (C) 2002 Prfecture du Bas-Rhin, France
# Copyright (C) 2002 Adelux <http://adequat.c2a.fr>
# Author: Florent Guillaume <mailto:fg@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: VersionManager.py,v 1.18 2003/07/14 21:40:49 fermigier Exp $

from zLOG import LOG, DEBUG

from Globals import InitializeClass
from AccessControl import getSecurityManager
from AccessControl import ClassSecurityInfo
from AccessControl import Unauthorized
from Acquisition import aq_base, aq_inner, aq_parent

from OFS.Folder import Folder

from Products.CMFCore.CMFCorePermissions import \
    ModifyPortalContent, AddPortalContent
from Products.CMFCore.CMFCorePermissions import AddPortalContent
from Products.CMFCore.utils import getToolByName
from Products.CMFDefault.DiscussionTool import DiscussionNotAllowed

from MiscTools import cut_and_paste
from MiscTools import copy_and_paste
from MiscTools import rename
from MiscTools import id_in_folder
from MiscTools import modifyRolesForPermission

from VersionCatalogTool import VersionCatalogTool


archives_id = 'archives'
drafts_id = 'drafts'
try:
    from Products.NuxCPS import isI18N
except:
    isI18N = 0

if isI18N:
    LOG('VersionManager', DEBUG, 'Using i18n drafts/archives titles')
    archive_title = '_str_Archives_'
    drafts_title = '_str_Drafts_'
else:
    LOG('VersionManager', DEBUG, 'Using local drafts/archives titles')
    archive_title = 'Archives'
    drafts_title = 'Version de travail'


class VersionManager:
    """
    This product requires the object to be associated with the
    'version_support' workflow definition.
    version attributes may be moved away from objectIds in the next release
        if not, ids (archives,drafts) will be changed to avoid collisions
    """
    security = ClassSecurityInfo()

    security.declarePublic('get_stable')
    def get_stable(self):
        """Returns the stable version of this document."""
        if self.is_stable():
            return aq_inner(self)
        else:
            return aq_parent(aq_inner(aq_parent(aq_inner(self))))

    security.declarePublic('is_locked')
    def is_locked(self):
        """Is the current document locked (no modifications allowed
        until a checkin)."""
        return getattr(aq_base(self), '_version_locked', 0)

    security.declarePublic('is_stable')
    def is_stable(self):
        """Is the current document the stable version."""
        return getattr(aq_base(self), '_stable', 1)

    security.declarePublic('is_archive')
    def is_archive(self):
        """Is the current document an archived version."""
        return getattr(aq_base(self), '_archive', 0)

    security.declarePublic('is_draft')
    def is_draft(self):
        """Is the current document a draft version."""
        if self.is_stable() or self.is_archive():
            return 0
        container = aq_parent(aq_inner(self))
        if container.getId() != drafts_id: # paranoid
            return 0
        return 1

    security.declarePublic('is_checkin_allowed')
    def is_checkin_allowed(self):
        """Returns true if the document can be checked in"""
        # Make sure this is a draft
        if not self.is_draft():
            return 0

        # Get stable's publication state and history
        stable = aq_parent(aq_inner(aq_parent(aq_inner(self))))
        pub = getattr(aq_base(stable), '_public_categories', [])
        pend = getattr(aq_base(stable), '_pending_categories', [])
        if pub or pend:
            # stable is pending or published, so only Managers
            # can check in
            # TODO: reviewers can also check-in
            roles = getSecurityManager().getUser().getRolesInContext(stable)
            if 'Manager' not in roles:
                return 0

        return 1

    security.declarePublic('is_checkout_allowed')
    def is_checkout_allowed(self):
        """Returns true if the document can be checked out."""
        # Make sure this is stable or archive
        if self.is_archive():
            stable = aq_parent(aq_inner(aq_parent(aq_inner(self))))
        elif self.is_stable():
            stable = self
        else:
            return 0

        # Make sure stable isn't locked
        if stable.is_locked():
            return 0

        return 1


    security.declarePrivate('become_locked')
    def become_locked(self):
        """Locks a stable version to forbid new drafts and modifications."""

        self._stable = 1
        self._archive = 0
        self._version_locked = 1

        # This has a guard that checks ModifyPortalContent, so do it
        # before blocking permissions
        getToolByName(self, 'portal_workflow').doActionFor(self, 'becomes_locked')
        self.reindexSubObjects(['version_state']) # reindex comments

        # Permission mapping
        self.set_permissions()
        #self.manage_permission(AddPortalContent,self.get_modifier_roles(),0)


    security.declarePrivate('become_stable')
    def become_stable(self,):
        """Become stable after being a draft or locked."""
        #self.unindexObject()

        # (do not use is_draft() as that one checks that we're in the
        # draft subdirectory)
        was_draft = not (self.is_stable() or self.is_archive())

        self._stable = 1
        self._archive = 0
        self._version_locked = 0

        # Restore workflows if it was a draft with workflows disabled
        if was_draft and hasattr(aq_base(self), 'setWorkflowSetChain'):
            self.setWorkflowSetChain(None)

        try:
            getToolByName(self,'portal_workflow').doActionFor(self,'becomes_stable')
            self.reindexSubObjects(['version_state']) # reindex comments
        except:
            LOG('VersionManager', DEBUG, "can't do becomes_stable")
            # XXX because can be called from __init__ where CMF hasn't
            # yet set portal_type, the workflows may be wrong
            pass

        # Permission mapping
        self.set_permissions()
        #self.manage_permission(AddPortalContent,self.get_modifier_roles(),0)

        #self.indexObject()

    security.declarePrivate('become_draft')
    def become_draft(self):
        """Become a draft after being stable or an archive."""
        #self.unindexObject()

        self._stable = 0
        self._archive = 0
        self._version_locked = 0 # no real meaning for drafts

        base_self = aq_base(self)

        # Get rid of stable version attributes
        if hasattr(base_self, drafts_id):
            self.manage_delObjects(ids=[drafts_id])
        if hasattr(base_self, archives_id):
            self.manage_delObjects(ids=[archives_id])

        # Get rid of discussion
        if hasattr(base_self, 'talkback'):
            talkback = self.talkback
            replies = talkback.getReplies()
            ids = [r.getId() for r in replies]
            for id in ids:
                try:
                    talkback.deleteReply(id)
                except KeyError:
                    LOG('VersionManager', DEBUG, 
                        "couldn't delete reply %s" % id)
                    pass # XXX ?
            del self.talkback
        # Recreate an empty talkback (to avoid a CMF bug with acquisition)
        disctool = getToolByName(self, 'portal_discussion', None)
        if disctool is not None:
            try:
                disctool._createDiscussionFor(self)
            except DiscussionNotAllowed:
                pass

        wftool = getToolByName(self, 'portal_workflow')

        # be sure PendingCategories and PublicCategories are cleared
        if 'publication_workflow' in wftool.getChainFor(self):
            self._public_categories = []
            self._pending_categories = []
            wftool.doActionFor(self, 'refresh')

        # Reset workflows. For now, workflows in setchain are lost,
        # workflows in addchain will be restored in become_stable
        # perhaps in future release set workflows will be remembered.
        if hasattr(base_self, 'setWorkflowSetChain'):
            self.setWorkflowSetChain(['version_support'])

        # This will also set correct publication permissions. XXX ???
        # Change state in the workflow.
        wftool.doActionFor(self, 'becomes_draft')
        self.reindexSubObjects(['version_state']) # reindex comments

        # Permission mapping
        self.set_permissions()

        #self.indexObject()


    security.declarePrivate('become_archive')
    def become_archive(self):
        """Become an archive after being a stable version."""
        #self.unindexObject()

        self._stable = 0
        self._archive = 1
        self._version_locked = 0 # no real meaning for archives

        base_self = aq_base(self)

        wftool = getToolByName(self, 'portal_workflow')

        # be sure PendingCategories and PublicCategories are cleared
        if 'publication_workflow' in wftool.getChainFor(self):
            self._public_categories = []
            self._pending_categories = []
            wftool.doActionFor(self, 'refresh')

        # Reset workflow aware objects
        if hasattr(base_self, 'setWorkflowSetChain'):
            self.setWorkflowSetChain(['version_support'])

        wftool.doActionFor(self, 'becomes_archive')
        self.reindexSubObjects(['version_state']) # reindex comments

        # Permission mapping
        self.set_permissions()

        #self.indexObject()

    def get_modifier_roles(self):
        """Get the modifier roles for this document."""
        base_self = aq_base(self)
        locked = getattr(base_self, '_version_locked', 0)
        archive = getattr(base_self, '_archive', 0)
        pub = getattr(base_self, '_public_categories', [])
        pend = getattr(base_self, '_pending_categories', [])
        if locked or archive or pub or pend:
            return ('Manager', )
        else:
            return ('Manager', 'WorkgroupManager', 'WorkgroupMember')

    def set_permissions(self):
        """Sets adequate permissions."""
        modifyRolesForPermission(self, ModifyPortalContent,
                                 self.get_modifier_roles())

##     def get_viewer_roles(self):
##         """Get the viewer roles for this document."""
##         return ('Manager', 'WorkgroupManager', 'WorkgroupMember', 'WorkgroupVisitor')

##     def set_private_permissions(self, ob):
##         """Sets private permissions"""
##         modifyRolesForPermission(ob, AccessContentsInformation,
##                                  self.get_viewer_roles())
##         modifyRolesForPermission(ob, View,
##                                  self.get_viewer_roles())
##         modifyRolesForPermission(ob, ModifyPortalContent,
##                                  self.get_modifier_roles())

    security.declarePrivate('container_attribute')
    def container_attribute(self, id, title):
        """XXX: move out to aux? for now this adds the attribute into
        objectIds.  this may change in future versions """
        ob = aq_base(self)
        if not hasattr(ob,id):
            attr = Folder()
            attr._setId(id)
            attr.isPrincipiaFolderish = 0
            self._setObject(id,attr)
            attr = self._getOb(id)
            attr.title = title
            # set permissions so that archives/drafts is private
            #self.set_private_permissions(attr)

    security.declarePrivate('redirect_user')
    def redirect_user(self, REQUEST):
        """ """
        if not REQUEST:
            return self

        if self.isPrincipiaFolderish:
            suffix = ''
        else:
            suffix = '/view'
        REQUEST.RESPONSE.redirect(self.absolute_url() + suffix)
        return None

    #
    # Transitions called by use actions
    #

    security.declareProtected(AddPortalContent, 'check_out')
    def check_out(self, REQUEST=None):
        """Creates a working version, and lock the stable version.
        Can be applied to a stable version or an archive."""

        if self.is_archive():
            stable = aq_parent(aq_inner(aq_parent(aq_inner(self))))
        elif self.is_stable():
            if self.is_locked():
                LOG('VersionManager', DEBUG, 'doc is locked')
                raise Unauthorized, "This version is marked as locked"
            stable = self
        else:
            LOG('VersionManager', DEBUG, 'doc is not stable or archive')
            raise Unauthorized, "This version is not marked as stable or archive"

        # Ensures drafts & archives exist
        stable.container_attribute(archives_id, archive_title)
        stable.container_attribute(drafts_id, drafts_title)
        drafts = getattr(stable, drafts_id)

        # Make copy in drafts
        stableid = stable.getId()
        draft_id = 'draft1'
        copy_and_paste(drafts, self, paste_as=draft_id)
        draft = getattr(drafts, draft_id)

        # Perform necessary state changes
        stable.become_locked()
        draft.become_draft()

        return draft.redirect_user(REQUEST)

    security.declareProtected(ModifyPortalContent, 'check_in')
    def check_in(self, REQUEST=None):
        """Creates an archived version from the stable document,
        and sets the current document as the new stable document."""

        if not self.is_checkin_allowed():
            raise Unauthorized, "You're not in a draft or you can't check in"

        stable = aq_parent(aq_inner(aq_parent(aq_inner(self))))
        stableid = stable.getId()
        draftid = self.getId()

        # Get stable's publication state and history
        pub = getattr(aq_base(stable), '_public_categories', [])
        pend = getattr(aq_base(stable), '_pending_categories', [])
        pubhist = getattr(aq_base(stable), 'workflow_history', {}).get('publication_workflow', ())

        tmpid = 'temporary_buffer_for_version_%s' % stableid

        #move old stable aside
        container = aq_parent(aq_inner(stable))
        rename(stable, tmpid)
        old_stable = container[tmpid]
        archives = old_stable[archives_id]
        drafts = old_stable[drafts_id]

        #move draft as new stable
        draft = drafts[draftid]
        cut_and_paste(container, draft, paste_as=stableid)
        new_stable = container[stableid]

        #move archives & drafts in new stable
        cut_and_paste(new_stable, archives)
        archives = new_stable[archives_id]
        cut_and_paste(new_stable, drafts)
        drafts = new_stable[drafts_id]

        #move old stable into archives
        archive_id = id_in_folder(archives, 'v', start_with=0)
        cut_and_paste(archives, old_stable, paste_as=archive_id)
        archive = archives[archive_id]

        # perform necessary state changes
        archive.become_archive()
        new_stable.become_stable()

        # set new stable's publication state
        new_stable._public_categories = pub
        new_stable._pending_categories = pend
        new_stable.workflow_history['publication_workflow'] = pubhist
        getToolByName(self, 'portal_workflow').doActionFor(new_stable, 'refresh')
        new_stable.reindexSubObjects(['version_state']) # reindex comments

        return new_stable.redirect_user(REQUEST)


    security.declareProtected(ModifyPortalContent,'abandon_draft')
    def abandon_draft(self, REQUEST=None):
        """
        Abandons a draft.
        """
        if not self.is_draft():
            raise Unauthorized, "You're not in a draft."

        drafts = aq_parent(aq_inner(self))
        stable = aq_parent(aq_inner(drafts))

        draftid = self.getId()
        drafts.manage_delObjects(ids=[draftid])

        # Perform necessary state changes
        stable.become_stable()

        return stable.redirect_user(REQUEST)

    security.declarePrivate('reindexSubObjects')
    def reindexSubObjects(self, idxs):
        """Reindex indexes in subobjects.
        """
        catalog = getToolByName(self, 'portal_catalog', None)
        if catalog is not None:
            path = '/'.join(self.getPhysicalPath())
            for brain in catalog.searchResults(path=path):
                ob = brain.getObject()
                if ob is None:
                    # Ignore old references to deleted objects.
                    continue
                s = getattr(ob, '_p_changed', 0)
                catalog.reindexObject(ob, idxs=idxs)
                if s is None: ob._p_deactivate()
            # Reindex the object itself, as the PathIndex only gave us
            # the descendants.
            self.reindexObject(idxs=idxs)

##     def get_catalog(self):
##         """Get the adequate catalog for the current object."""
##         if self.is_stable():
##             return getToolByName(self, 'portal_catalog', None)
##         else:
##             return getToolByName(self, 'portal_version_catalog', None)
##
##     def indexObject(self):
##         catalog = self.get_catalog()
##         if catalog is not None:
##             catalog.indexObject(self)
##
##     security.declareProtected(ModifyPortalContent, 'unindexObject')
##     def unindexObject(self):
##         catalog = self.get_catalog()
##         if catalog is not None:
##             catalog.unindexObject(self)
##
##     security.declareProtected(ModifyPortalContent, 'reindexObject')
##     def reindexObject(self, idxs=[]):
##         if idxs == []:
##             # Update the modification date.
##             if hasattr(aq_base(self), 'notifyModified'):
##                 self.notifyModified()
##         catalog = self.get_catalog()
##         if catalog is not None:
##             catalog.reindexObject(self, idxs=idxs)


InitializeClass(VersionManager)
